/*
 * Back-end main header
 * This is the most important file for you to understand the back-end.
 *
 * Copyright INOUE Seiichiro <inoue@ainet.or.jp>, licensed under the GPL.
 */
#ifndef __GDIFF_GDIFF_H__
#define __GDIFF_GDIFF_H__

#include <time.h> /* for time_t */
#include <glib.h>

#ifdef __cplusplus 
extern "C" {
#endif /* __cplusplus */

/*
 * General description of data structures
 *
 * DiffDir is allocated for one session of comparing(diff) two files
 * or two directories.
 * DiffDir manages a linked list of DiffFiles.
 * Each node(DiffFiles) is allocated for two different (regular)files
 * under the directories.
 * DiffFiles has a general information of two files,
 * such as filename or the number of lines.
 * In addition, DiffFiles manages a linked list of DiffLines.
 * Each node(DiffLines) is allocated for different part(=lines),
 * which has the type(DiffType) and the line numbers of the difference.
 *
 * DiffDir and DiffFiles are managed by reference count.
 * The initial value is zero, but the creator usually owns it and increments
 * the count.
 * DiffDir is typically owned by GUI directory view.
 * On the other hand, DiffFiles is owned by DiffDir and GUI file views.
 */

/* Forward declarations of types. */
typedef struct _DiffDir DiffDir;
typedef struct _DiffFiles DiffFiles;
typedef struct _FileInfo FileInfo;
typedef struct _DiffLines DiffLines;
typedef struct _DiffBeginEnd DiffBeginEnd;

/*
 * Constant number
 */
/* IMPORTANT. These numbers are used in DiffLines structure.
 * Basically, they represent type for each differences.
 * In addition, used as the return value from parse_diff_line() */
typedef enum {
	F1ONLY,	/* Lines taken from just the first file. (known as DELETE) */
	F2ONLY,	/* Lines taken from just the second file. (known as ADD) */
	CHANGE,
	/* The followings are used for the return value from parse_diff_line() */
	FINDFILE,
	FINDBINFILE,
	IGNORE,
	ERROR
} DiffType;


#define NUM_COMPARE_FILES		2
typedef enum {
	FIRST_FILE = 0,
	SECOND_FILE = 1
} WhichFile;

/* Status of two files.
   I'm afraid this might be confused with DiffType above.
   DiffType exists for each different parts(lines).
   On the other hand, DiffStatus is related to the files(directories) compared.
   These shifted bits are used in row_hide_stat_mask of DViewPref(see properties.h). */
typedef enum {
	BINARY_FILES		= (1<<0),
	ONLY_FILE1_EXISTS	= (1<<1),
	ONLY_FILE2_EXISTS	= (1<<2),
	DIFFERENT_FILES		= (1<<3),
	IDENTICAL_FILES		= (1<<4),
	DIRECTORIES			= (1<<5)
} FilesStatus;

typedef enum {
	MOD_MODIFIED,
	MOD_NOMODIFIED,
	MOD_DONTCARE,
	MOD_ERROR
} FileModStatue;

	
/*
 * Data structure definitions
 */

/* Directory information data
 * Note: Comparing two regular files is dealt with as
 * one special case of comparing two directories.
 * - When comparing directories, the first node of dfiles_list
 *   has the directories info, and the following nodes has each files info.
 * - When comparing two regular files, the first node has two files info. */
struct _DiffDir {
	int ref_count;			/* Reference count */
	GSList *dfiles_list;	/* Singly linked list of DiffFiles */
};


/* Compared files information data
 * Some special cases as follows,
 * - dlines_list==NULL implies two files are identical.
 * - One of fileinfo[]'s fname=="" implies that
 *   a file exists only in one directory.
 *
 * struct _FileInfo
 * Two possible cases,
 * 1)directory info. 2)file info. */
struct _DiffFiles {
	int ref_count;		/* Reference count */
	struct _FileInfo {
		char *fname;	/* file name. "" in non-exist case */
		int nlines;		/* Number of lines */
		char *buf;		/* (mmap'd)buffer */
		int lenb;		/* Length(Byte) */
		gboolean f_dir;	/* TRUE if directory */
		time_t mtime;	/* time of last modification */
		time_t ctime;	/* time of last change */
	} fileinfo[NUM_COMPARE_FILES];
	GList *dlines_list;/* Doubly linked list of DiffLines */
	gboolean binary;
};


/* Diff lines information data
 * Keeps difftype and the line numbers. */
struct _DiffLines {
	DiffType difftype;
	struct _DiffBeginEnd {
		/* these are line numbers. */
		int begin;
		int end;
	} between[NUM_COMPARE_FILES];
};


/* Function type definitions */
typedef const GList* (*DiffFindFunc)(const DiffFiles *dfiles, const GList *cur_dlines_node);
	

/* Global variable declarations */
/* None */


/* Global function declarations */
/* diffdata.c */
extern DiffDir* diffdir_new(const char *fname1, const char *fname2, const char *args);
extern void diffdir_ref(DiffDir *diffdir);
extern void diffdir_unref(DiffDir *diffdir);
extern DiffFiles* diffdir_add_dfiles(DiffDir *diffdir, const char *fname1, const char *fname2);
extern void diffdir_remove_dfiles(DiffDir *diffdir, DiffFiles *dfiles);

extern void dfiles_ref(DiffFiles *dfiles);
extern void dfiles_unref(DiffFiles *dfiles);
extern const FileInfo* dfiles_get_fileinfo(DiffFiles *dfiles, WhichFile n, gboolean f_mmap);
extern void dfiles_add_dlines(DiffFiles* dfiles, DiffType dtype, int f1n1, int f1n2, int f2n1, int f2n2);


/* rundiff.c */
extern void run_diff(DiffDir *diffdir, const char *filename1, const char *filename2, const char *args, DiffFiles *cur_files);


/* diffmisc.c */
extern FilesStatus dfiles_get_status(DiffFiles *dfiles);
extern FileModStatue dfiles_has_file_modified(DiffFiles *dfiles);
extern int dfiles_calc_total_nlines(const DiffFiles *dfiles, WhichFile n);
extern int dfiles_get_max_nlines(DiffFiles *dfiles);

/* DiffFindFunc functions */
extern const GList* dfiles_get_curl(const DiffFiles *dfiles, const GList *cur_dlines_node);
extern const GList* dfiles_get_firstl(const DiffFiles *dfiles, const GList *cur_dlines_node);
extern const GList* dfiles_get_lastl(const DiffFiles *dfiles, const GList *cur_dlines_node);
extern const GList* dfiles_find_nextl(const DiffFiles *dfiles, const GList *cur_dlines_node);
extern const GList* dfiles_find_prevl(const DiffFiles *dfiles, const GList *cur_dlines_node);

extern const GList* dfiles_find_rel_curl(const DiffFiles *dfiles, WhichFile n, int cur_line);
extern const GList* dfiles_find_rel_nextl(const DiffFiles *dfiles, WhichFile n, int cur_line);
extern const GList* dfiles_find_rel_prevl(const DiffFiles *dfiles, WhichFile n, int cur_line);

extern const GList* dfiles_find_includel(const DiffFiles *dfiles, WhichFile n, int cur_line);

#ifdef __cplusplus
} 
#endif /* __cplusplus */
 
#endif /* __GDIFF_GDIFF_H__ */
