/*
 * GUI file selection module
 * Mainly, handling GtkFileSelection widget.
 *
 * Copyright INOUE Seiichiro <inoue@ainet.or.jp>, licensed under the GPL.
 */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <sys/types.h>
#include <sys/stat.h>
#include <gnome.h>
#include "diff.h"
#include "misc.h"
#include "gui.h"
#include "gdwin.h"
#include "dirview.h"
#include "onepaneview.h"
#include "twopaneview.h"
#include "mergeview.h"


/* Note:
 * File dialog box has its internal state,
 * because it has to accept two file names.
 */

/* Constant strings */
static const char FIRST_FILE_SEL_CAPTION[] = N_("First file selection");
static const char SECOND_FILE_SEL_CAPTION[]	= N_("Second file selection");

/* Key for gtk_object_[get|set]_data() */
#define FILEDIALOGDATA_KEY	"fddkey"

/* Data structure definitions */
/* Internal data in GtkFileSelection instance.
 * Kept by gtk_object_set_data() and gtk_object_get_data(). */
typedef struct {
	/* For reference */
	GDiffWindow *gdwin;

	/* Dialog box state */
	gboolean is_directory;
	char *fname1;
	GtkLabel *fname1_label;
	WhichFile file_count;
} FileDialogData;


/* Private function declarations */
static void files_selection_dialog(GDiffWindow *gdwin);
static void file_selection_ok(GtkWidget *w, gpointer data);
static void file_selection_cancel(GtkWidget *w, gpointer data);
static void dir_toggle(GtkWidget *w, gpointer data);

static FileDialogData* filedialogdata_new(GDiffWindow *gdwin, GtkLabel *label);
static void filedialogdata_reset(FileDialogData *fdd);
static void filedialogdata_update(FileDialogData *fdd, const char *fname1);


/**
 * filesel_open:
 * An entry point to open two files(directories).
 * At present, called only from [menu]-[File]-[Open].
 **/
void
filesel_open(GDiffWindow *gdwin)
{
	files_selection_dialog(gdwin);
}


/* ---The followings are private functions--- */
/**
 * files_selection_dialog:
 * If the instance of dialog has not created yet, create it and show it.
 * If it exists and has been hidden, just show it.
 * If it has been shown, just raise it.
 **/
static void
files_selection_dialog(GDiffWindow *gdwin)
{
	FileDialogData *fdd;
	GtkWidget *filesel;
	GtkWidget *checkbutton;
	GtkWidget *label;

	/* The dialog has already been created */
	if (gdwin->filesel) {
		if (!GTK_WIDGET_VISIBLE(gdwin->filesel)) {
			fdd = gtk_object_get_data(GTK_OBJECT(gdwin->filesel), FILEDIALOGDATA_KEY);
			
			if (fdd->file_count == FIRST_FILE)
				gtk_window_set_title(GTK_WINDOW(gdwin->filesel),
									 _(FIRST_FILE_SEL_CAPTION));
			/* Not necessary, because I have reset all states at cancellation.
			else if (fdd->file_count == SECOND_FILE)
				gtk_window_set_title(GTK_WINDOW(fs), SECOND_FILE_SEL_CAPTION);
				*/
			gtk_widget_show(GTK_WIDGET(gdwin->filesel));
		} else {
			gdk_window_raise(GTK_WIDGET(gdwin->filesel)->window);
		}
		return;
	}

	/* Create GtkFileSelection dialog */
	filesel = gtk_file_selection_new(_(FIRST_FILE_SEL_CAPTION));
	gtk_file_selection_hide_fileop_buttons(GTK_FILE_SELECTION(filesel));
	gtk_window_set_position(GTK_WINDOW(filesel), GTK_WIN_POS_MOUSE);
	gdwin->filesel = GTK_FILE_SELECTION(filesel);

	/* Label to show the first file name, while you're selecting a second file. */
	label = gtk_label_new("");
	gtk_box_pack_start(GTK_BOX(GTK_FILE_SELECTION(filesel)->action_area), 
					   label, FALSE, FALSE, 0);

	/* Set FileDialogData as internal data of file dialog widget */
	fdd = filedialogdata_new(gdwin, GTK_LABEL(label));
	gtk_object_set_data(GTK_OBJECT(filesel), FILEDIALOGDATA_KEY, fdd);
	
	gtk_signal_connect(GTK_OBJECT(filesel), "destroy",
					   GTK_SIGNAL_FUNC(gtk_widget_destroyed), &gdwin->filesel);
	gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(filesel)->ok_button),
					   "clicked", GTK_SIGNAL_FUNC(file_selection_ok), filesel);
	gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(filesel)->cancel_button),
					   "clicked", GTK_SIGNAL_FUNC(file_selection_cancel), filesel);

	/* Checkbutton to specify directory selection */
	checkbutton = gtk_check_button_new_with_label(_("Compare directories"));
	/* XXX:what effect? gtk_toggle_button_set_mode(GTK_TOGGLE_BUTTON(checkbutton), TRUE);*/
	gtk_signal_connect(GTK_OBJECT(checkbutton),
					   "toggled", GTK_SIGNAL_FUNC(dir_toggle), filesel);
	gtk_box_pack_start(GTK_BOX(GTK_FILE_SELECTION(filesel)->action_area), 
					   checkbutton, FALSE, FALSE, 0);
	gtk_widget_show(checkbutton);
	
    gtk_widget_show(filesel);
}

/**
 * file_selection_ok:
 * The behavior depends on the internal state.
 * In case first file selected, change its state for second file.
 * In case second file selected, do the job.
 **/
static void
file_selection_ok(GtkWidget *w, gpointer data)
{
	GtkFileSelection *fs = data;
	FileDialogData *fdd;
	char *fname;
	FileType ft;

	fdd = gtk_object_get_data(GTK_OBJECT(fs), FILEDIALOGDATA_KEY);
	fname = gtk_file_selection_get_filename(fs);
	ft = check_filetype(fname);
	if (ft == OTHERFILE) {
		g_message(_("Wrong file type\n"));
		return;
	} else if (ft == REGULARFILE && fdd->is_directory) {
		g_message(_("You have to select a directory. Or don't check the option button.\n"));
		return;
	} else if (ft == DIRECTORY && !(fdd->is_directory)) {
		g_message(_("You have to select a (regular)file. Or check the option button.\n"));
		return;
	}
	
	if (fdd->file_count == FIRST_FILE) {
		filedialogdata_update(fdd, fname);
		gtk_window_set_title(GTK_WINDOW(fs), _(SECOND_FILE_SEL_CAPTION));
		gtk_file_selection_set_filename(GTK_FILE_SELECTION(fs), "");
	} else if (fdd->file_count == SECOND_FILE) {
		GDiffWindow *gdwin;
		GtkWidget *new_view = NULL;
		DiffDir *diffdir;
		char *fname1;
		char *fname2;
		
		gdwin = fdd->gdwin;
		fname1 = fdd->fname1;
		fname2 = fname;

		diffdir = diffdir_new(fname1, fname2, g_pref.diff_args);
		if (fdd->is_directory) {
			new_view = gdiff_dirview_new(diffdir);
		} else {
			DiffFiles *dfiles = g_slist_nth_data(diffdir->dfiles_list, 0);
			if (g_pref.fvpref.view_type == ONEPANE_VIEW)
				new_view = gdiff_onepview_new(diffdir, dfiles, FALSE);
			else if (g_pref.fvpref.view_type == TWOPANE_VIEW)
				new_view = gdiff_twopview_new(diffdir, dfiles, FALSE);
			else if (g_pref.fvpref.view_type == MERGE_VIEW)
				new_view = gdiff_mergeview_new(diffdir, dfiles, FALSE);
			else
				g_assert_not_reached();
		}

		filedialogdata_reset(fdd);
		gtk_widget_hide(GTK_WIDGET(fs));

		/* Do the job */
		gdwin_add_view(gdwin, new_view);
	} else
		g_assert_not_reached();
}

/**
 * file_selection_cancel:
 * Reset all internal states.
 **/
static void
file_selection_cancel(GtkWidget *w, gpointer data)
{
	GtkFileSelection *fs = data;
	FileDialogData *fdd;

	fdd = gtk_object_get_data(GTK_OBJECT(fs), FILEDIALOGDATA_KEY);
	filedialogdata_reset(fdd);
	gtk_widget_hide(GTK_WIDGET(fs));
}

/**
 * dir_toggle:
 * Toggle whether to select two directories or two files.
 * If a user has already selected the first one, don't change the mode.
 **/
static void
dir_toggle(GtkWidget *w, gpointer data)
{
	GtkFileSelection *fs = data;
	FileDialogData *fdd;

	fdd = gtk_object_get_data(GTK_OBJECT(fs), FILEDIALOGDATA_KEY);
		
	if (fdd->file_count == FIRST_FILE) {
		fdd->is_directory = !(fdd->is_directory);
		gtk_widget_set_sensitive(fs->file_list, !(fdd->is_directory));
	} else /* Don't change */
		GTK_TOGGLE_BUTTON(w)->active = !GTK_TOGGLE_BUTTON(w)->active;/* Is this a proper way ?*/
}


/* Routines for dialog box's own private data, i.e. FileDialogData */
/**
 * filedialogdata_new:
 * Allocate FileDialogData, intialize it and return its pointer.
 * Input:
 * GDiffWindow *gdwin; kept by FileDialogData.
 * GtkLabel *label; kept by FileDialogData.
 * Output:
 * Return value; allocated FileDialogData pointer.
 **/
static FileDialogData*
filedialogdata_new(GDiffWindow *gdwin, GtkLabel *label)
{
	FileDialogData *fdd;

	fdd = g_new(FileDialogData, 1);

	fdd->gdwin = gdwin;
	fdd->is_directory = FALSE;
	fdd->fname1 = NULL;
	fdd->fname1_label = label;
	fdd->file_count = FIRST_FILE;

	return fdd;
}

/**
 * filedialogdata_reset:
 * Reset FileDialogData.
 * Called when a user closes a dialog while he has selected a first file,
 * but hasn't selected a second file.
 * Input:
 * FileDialogData *fdd;
 * Output:
 * FileDialogData *fdd; updated.
 **/
static void
filedialogdata_reset(FileDialogData *fdd)
{
	if (fdd->fname1 == NULL)
		return;
	/* Not change: fdd->is_directory */
	g_free(fdd->fname1);
	fdd->fname1 = NULL;
	gtk_widget_hide(GTK_WIDGET(fdd->fname1_label));
	fdd->file_count = FIRST_FILE;
}

/**
 * filedialogdata_update:
 * Update FileDialogData.
 * Called when a user selects a first file.
 * Input:
 * FileDialogData *fdd;
 * const char *fname1; Selected first file name.
 * Output:
 * FileDialogData *fdd; updated.
 **/
static void
filedialogdata_update(FileDialogData *fdd, const char *fname1)
{
	fdd->fname1 = g_strdup(fname1);
	gtk_label_set_text(fdd->fname1_label, fname1);
	gtk_widget_show(GTK_WIDGET(fdd->fname1_label));
	fdd->file_count = SECOND_FILE;
}
