/*
 * GdiffOverview widget module
 *
 * Copyright INOUE Seiichiro <inoue@ainet.or.jp>, licensed under the GPL.
 */
#include <gtk/gtksignal.h>
#include "gdiffoverview.h"
#include "style.h"


/* Private function declarations */
static void gdiff_overview_class_init(GdiffOverviewClass *klass);
static void gdiff_overview_init(GdiffOverview *overview);

static void gdiff_overview_draw_lines(GtkWidget *widget, GdkEventExpose *event, gpointer data);


static GtkHBoxClass *parent_class = NULL;

GtkType
gdiff_overview_get_type(void)
{
  static GtkType overview_type = 0;

  if (!overview_type) {
      static const GtkTypeInfo overview_info = {
		  "GdiffOverview",
		  sizeof(GdiffOverview),
		  sizeof(GdiffOverviewClass),
		  (GtkClassInitFunc)gdiff_overview_class_init,
		  (GtkObjectInitFunc)gdiff_overview_init,
		  /* reserved_1 */ NULL,
		  /* reserved_2 */ NULL,
		  (GtkClassInitFunc)NULL,
      };
      overview_type = gtk_type_unique(GTK_TYPE_HBOX, &overview_info);
  }
  
  return overview_type;
}

static void
gdiff_overview_class_init(GdiffOverviewClass *klass)
{
	parent_class = gtk_type_class(GTK_TYPE_HBOX);
}

static void
gdiff_overview_init(GdiffOverview *overview)
{
	GtkHBox *hbox;
	GtkWidget *range1;
	GtkWidget *range2;
	GtkWidget *darea;

	hbox = GTK_HBOX(overview);
	
	range1 = gdiff_range_new(NULL);
	overview->range1 = GDIFF_RANGE(range1);
	gtk_widget_set_name(range1, "file1 overview");
	style_set_gdiff_range1(range1);
	gtk_widget_show(range1);

	range2 = gdiff_range_new(NULL);
	overview->range2 = GDIFF_RANGE(range2);
	gtk_widget_set_name(range2, "file2 overview");
	style_set_gdiff_range2(range2);
	gtk_widget_show(range2);

	darea = gtk_drawing_area_new();
	overview->darea = GTK_DRAWING_AREA(darea);
	gtk_widget_set_name(darea, "drawing overview");
	style_set_drawingarea(darea);
	gtk_signal_connect(GTK_OBJECT(darea), "expose_event",
					   GTK_SIGNAL_FUNC(gdiff_overview_draw_lines), overview);
	gtk_widget_show(darea);

	/* adjust sizes */
	gtk_drawing_area_size(GTK_DRAWING_AREA(darea), WIDTH_OVERVIEW, 0);
	gdiff_range_size(GDIFF_RANGE(range1), WIDTH_OVERVIEW, 0);
	gdiff_range_size(GDIFF_RANGE(range2), WIDTH_OVERVIEW, 0);

	gtk_box_pack_start(GTK_BOX(hbox), range1, FALSE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(hbox), darea, FALSE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(hbox), range2, FALSE, FALSE, 0);
}

GtkWidget*
gdiff_overview_new(GtkAdjustment *adj1, GtkAdjustment *adj2)
{
	GdiffOverview *overview;

	overview = gtk_type_new(GDIFF_TYPE_OVERVIEW);
	gdiff_overview_set_adjustment1(overview, adj1);
	gdiff_overview_set_adjustment2(overview, adj2);
	
	return GTK_WIDGET(overview);
}


/** Interfaces **/
/**
 * gdiff_overview_set_adjustment1:
 **/
void
gdiff_overview_set_adjustment1(GdiffOverview *overview, GtkAdjustment *adj)
{
	g_return_if_fail(overview != NULL);
	g_return_if_fail(GDIFF_IS_OVERVIEW(overview));

	gdiff_range_set_adjustment(overview->range1, adj);
}

/**
 * gdiff_overview_set_adjustment2:
 **/
void
gdiff_overview_set_adjustment2(GdiffOverview *overview, GtkAdjustment *adj)
{
	g_return_if_fail(overview != NULL);
	g_return_if_fail(GDIFF_IS_OVERVIEW(overview));

	gdiff_range_set_adjustment(overview->range2, adj);
}

/**
 * gdiff_overview_insert_paintrange:
 * Internally, set it for GdiffRange widgets.
 **/
void
gdiff_overview_insert_paintrange(GdiffOverview *overview, gdouble begin1, gdouble end1, gdouble begin2, gdouble end2)
{
	g_return_if_fail(overview != NULL);
	g_return_if_fail(GDIFF_IS_OVERVIEW(overview));

	gdiff_range_insert_paintrange(overview->range1, begin1, end1);
	gdiff_range_insert_paintrange(overview->range2, begin2, end2);
}

/**
 * gdiff_overview_set_color1:
 **/
void gdiff_overview_set_color1(GdiffOverview *overview, GdkColor *fg, GdkColor *bg)
{
	g_return_if_fail(overview != NULL);
	g_return_if_fail(GDIFF_IS_OVERVIEW(overview));

	if (fg)
		gdiff_range_set_foreground(overview->range1, fg);
	if (bg)
		gdiff_range_set_background(overview->range1, bg);
}

/**
 * gdiff_overview_set_color2:
 **/
void gdiff_overview_set_color2(GdiffOverview *overview, GdkColor *fg, GdkColor *bg)
{
	g_return_if_fail(overview != NULL);
	g_return_if_fail(GDIFF_IS_OVERVIEW(overview));

	if (fg)
		gdiff_range_set_foreground(overview->range2, fg);
	if (bg)
		gdiff_range_set_background(overview->range2, bg);
}


/** Internal functions **/
#define lines_intersect(y1, y2, top, bottom) \
	((y1 >= top && y1 <= bottom) || (y2 >= top && y2 <= bottom))

/**
 * gdiff_overview_draw_lines:
 * Using range values, which are kept in GdiffRange widgets,
 * draw lines on drawingarea widget.
 **/
static void
gdiff_overview_draw_lines(GtkWidget *widget, GdkEventExpose *event, gpointer data)
{
	GdiffOverview *overview = data;
	GdkWindow *w = widget->window;
	GdkGC *gc = widget->style->fg_gc[GTK_STATE_NORMAL];
	int darea_width;
	int darea_height;
	const GSList *list1 = overview->range1->range_list;
	const GSList *list2 = overview->range2->range_list;
	
	gdk_window_get_size(w, &darea_width, &darea_height);
	
	while (list1 && list2) {
		const PaintRange *pr1 = list1->data;
		const PaintRange *pr2 = list2->data;
		int y1, y2;
		
		y1 = darea_height * (pr1->begin + pr1->end) / 2;
		y2 = darea_height * (pr2->begin + pr2->end) / 2.0;
		
		if (lines_intersect(y1, y2,
							event->area.y, event->area.y+event->area.height)) {
			gdk_draw_line(w, gc, 0, y1, darea_width, y2);
		}
		list1 = g_slist_next(list1);
		list2 = g_slist_next(list2);
	}
}
