/*
 * Merge widget module
 *
 * This module is complicated, because there are many states.
 * The states can be described by the order of DispLines objects.
 *
 * Case:only basefile has the parts. There are two patterns.
 * DA_MARK			==insert(b)=>		DA_ONLY
 * DA_ONLY|DA_HIDE	<=remove(b)==
 *
 * Case:only otherfile has the parts. There are two patterns.
 * DA_O_ONLY		==insert(o)=>		DA_O_ONLY
 * DA_MARK			<=remove(o)==		DA_ONLY_O
 *
 * Case:both files have different parts. There are six patterns...
 * DA_MARK(*)		==insert(b)=>	DA_CHANGE(**)==insert(o)=>	DA_CHANGE(***)
 * DA_CHANGE|DA_HIDE<=remove(b)==				<=remove(o)==	DA_CHANGE_O
 *
 * DA_MARK(*)		==insert(o)=>	DA_CHANGE_O		==insert(b)=>	DA_CHANGE_O(****)
 * DA_CHANGE|DA_HIDE<=remove(o)==	DA_CHANGE|DA_HIDE<=remove(b)==	DA_CHANGE
 *
 * DA_CHANGE(***)	==remove(b)=>	DA_CHANGE|DA_HIDE	==remove(o)=>	DA_MARK(*)
 * DA_CHANGE_O		<=insert(b)==	DA_CHANGE_O							DA_CHANGE|DA_HIDE
 *
 * DA_CHANGE_O(****) ==remove(o)=>	DA_CHANGE(**)
 * DA_CHANGE
 *
 * Copyright INOUE Seiichiro <inoue@ainet.or.jp>, licensed under the GPL.
 */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <gtk/gtksignal.h>
#include <gtk/gtkvpaned.h>
#include "diff.h"
#include "gui.h"
#include "dtextmap.h"
#include "merge-widget.h"
#include "misc.h"
#include "gtktext-support.h"
#include "linenum.h"
#include "style.h"


#define IS_THIS_ONLY(whichfile, difftype)	\
	((whichfile == FIRST_FILE && difftype == F1ONLY)	\
	 || (whichfile == SECOND_FILE && difftype == F2ONLY))
#define IS_OTHER_ONLY(whichfile, difftype)	\
	((whichfile == FIRST_FILE && difftype == F2ONLY)	\
	 || (whichfile == SECOND_FILE && difftype == F1ONLY))
	
static WhichFile basefile = FIRST_FILE;
static WhichFile otherfile = SECOND_FILE;

static const char marktext[] = "$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$\n";/* character doesn't matter */
static const int marktext_lenb = sizeof(marktext)-1;


/* Private function declarations */
static void gdiff_merge_class_init(GdiffMergeClass *klass);
static void gdiff_merge_init(GdiffMerge *merge);

static void gdiff_merge_display(GdiffBasePane *basepane);
static void gdiff_merge_show_linenum(GdiffBasePane *basepane, gboolean to_show);
static void gdiff_merge_select_dlines(GdiffBasePane *basepane, WhichFile whichfile, int ln);

static void guts_merge_display(GdiffMerge *merge);
static void draw_text(GdiffMerge *merge);
static void show_hide_numbers(GdiffMerge *merge, gboolean b_show);

static gint text_click_cb(GtkWidget *text, GdkEventButton *event, gpointer data);
static void insert_base(GdiffMerge *merge, const DiffLines *dlines);
static void insert_other(GdiffMerge *merge, const DiffLines *dlines);
static void remove_base(GdiffMerge *merge, const DiffLines *dlines);
static void remove_other(GdiffMerge *merge, const DiffLines *dlines);

static void remove_displn(DTextMap *dtmap, GtkWidget *text, DispLines *displn);

static void guts_output_file(GdiffMerge *merge, const char *filename);


static GdiffOnePaneClass *parent_class = NULL;


GtkType
gdiff_merge_get_type(void)
{
  static GtkType merge_type = 0;

  if (!merge_type) {
      static const GtkTypeInfo merge_info = {
		  "GdiffMerge",
		  sizeof(GdiffMerge),
		  sizeof(GdiffMergeClass),
		  (GtkClassInitFunc)gdiff_merge_class_init,
		  (GtkObjectInitFunc)gdiff_merge_init,
		  /* reserved_1 */ NULL,
		  /* reserved_2 */ NULL,
		  (GtkClassInitFunc)NULL,
      };
      merge_type = gtk_type_unique(GDIFF_TYPE_ONEPANE, &merge_info);
  }
  
  return merge_type;
}

static void
gdiff_merge_class_init(GdiffMergeClass *klass)
{
	GtkObjectClass *object_class;
	GdiffBasePaneClass *basepane_class;
	
	object_class = (GtkObjectClass*)klass;
	basepane_class = (GdiffBasePaneClass*)klass;
	parent_class = gtk_type_class(GDIFF_TYPE_ONEPANE);

	basepane_class->display = gdiff_merge_display;
	basepane_class->show_linenum = gdiff_merge_show_linenum;
	basepane_class->select_dlines = gdiff_merge_select_dlines;
}

static void
gdiff_merge_init(GdiffMerge *merge)
{
	GdiffOnePane *onepane;

	onepane = GDIFF_ONEPANE(merge);
	PANE_PREF(merge).view_type = MERGE_VIEW;
}

GtkWidget*
gdiff_merge_new(DiffDir *diffdir, DiffFiles *dfiles)
{
	GdiffMerge *merge;
	const FileInfo *fi1 = dfiles_get_fileinfo(dfiles, FIRST_FILE, TRUE);

	merge = gtk_type_new(GDIFF_TYPE_MERGE);

	_gdiff_basepane_set_backend(GDIFF_BASEPANE(merge), diffdir, dfiles);/*XXX*/
	
	/* Merge-pane internal data */
	GDIFF_ONEPANE(merge)->dtmap = dtmap_new(fi1->nlines);

	gtk_signal_connect_after(GTK_OBJECT(GDIFF_ONEPANE(merge)->text),
							 "button_press_event",
							 GTK_SIGNAL_FUNC(text_click_cb), merge);

	/* workaround */
	gtk_widget_ensure_style(GDIFF_ONEPANE(merge)->text);
	
	guts_merge_display(merge);
	
	return GTK_WIDGET(merge);
}


/** Interfaces **/
static void
gdiff_merge_display(GdiffBasePane *basepane)
{
	GdiffMerge *merge;

	g_return_if_fail(basepane != NULL);
	g_return_if_fail(GDIFF_IS_MERGE(basepane));

	merge = GDIFF_MERGE(basepane);

	guts_merge_display(merge);
}

static void
gdiff_merge_show_linenum(GdiffBasePane *basepane, gboolean to_show)
{
	GdiffMerge *merge;

	g_return_if_fail(basepane != NULL);
	g_return_if_fail(GDIFF_IS_MERGE(basepane));

	merge = GDIFF_MERGE(basepane);
	
	if (to_show != PANE_PREF(merge).show_line_num) {
		show_hide_numbers(merge, to_show);
		PANE_PREF(merge).show_line_num = to_show;
	}
}

static void
gdiff_merge_select_dlines(GdiffBasePane *basepane, WhichFile whichfile, int ln)
{
	GdiffMerge *merge;
	DiffFiles *dfiles;
	const GList *dlines_node;

	g_return_if_fail(basepane != NULL);
	g_return_if_fail(GDIFF_IS_MERGE(basepane));

	merge = GDIFF_MERGE(basepane);
	dfiles = PANE_DFILES(merge);

	dlines_node = dfiles_find_includel(dfiles, whichfile, ln);
	if (dlines_node == NULL && ln == 1)
		dlines_node = dfiles_find_rel_curl(dfiles, whichfile, 0);/* exception */
	
	if (dlines_node) {
		/* todo coloring */

		GDIFF_BASEPANE(merge)->cur_dlines_node = dlines_node;
		gtk_signal_emit_by_name(GTK_OBJECT(merge), "move_diff", MOVED_CUR_NOSCROLL);
	}
}


void
gdiff_merge_insert_basefile_all(GdiffMerge *merge)
{
	DiffFiles *dfiles;
	GList *node;/* node of DiffLines list */
	GtkWidget *text;
	
	g_return_if_fail(merge != NULL);
	g_return_if_fail(GDIFF_IS_MERGE(merge));

	dfiles = PANE_DFILES(merge);
	text = GDIFF_ONEPANE(merge)->text;

	gtext_freeze(text);
	for (node = dfiles->dlines_list; node; node = node->next) {
		const DiffLines *dlines = node->data;
	
		insert_base(merge, dlines);
	}
	gtext_thaw(text);
}

void
gdiff_merge_remove_basefile_all(GdiffMerge *merge)
{
	DiffFiles *dfiles;
	GList *node;/* node of DiffLines list */
	GtkWidget *text;
	
	g_return_if_fail(merge != NULL);
	g_return_if_fail(GDIFF_IS_MERGE(merge));

	dfiles = PANE_DFILES(merge);
	text = GDIFF_ONEPANE(merge)->text;

	gtext_freeze(text);
	for (node = dfiles->dlines_list; node; node = node->next) {
		const DiffLines *dlines = node->data;
	
		remove_base(merge, dlines);
	}
	gtext_thaw(text);
}

void
gdiff_merge_insert_otherfile_all(GdiffMerge *merge)
{
	DiffFiles *dfiles;
	GList *node;/* node of DiffLines list */
	GtkWidget *text;
	
	g_return_if_fail(merge != NULL);
	g_return_if_fail(GDIFF_IS_MERGE(merge));

	dfiles = PANE_DFILES(merge);
	text = GDIFF_ONEPANE(merge)->text;

	gtext_freeze(text);
	for (node = dfiles->dlines_list; node; node = node->next) {
		const DiffLines *dlines = node->data;
	
		insert_other(merge, dlines);
	}
	gtext_thaw(text);
}

void
gdiff_merge_remove_otherfile_all(GdiffMerge *merge)
{
	DiffFiles *dfiles;
	GList *node;/* node of DiffLines list */
	GtkWidget *text;
	
	g_return_if_fail(merge != NULL);
	g_return_if_fail(GDIFF_IS_MERGE(merge));

	dfiles = PANE_DFILES(merge);
	text = GDIFF_ONEPANE(merge)->text;

	gtext_freeze(text);
	for (node = dfiles->dlines_list; node; node = node->next) {
		const DiffLines *dlines = node->data;
	
		remove_other(merge, dlines);
	}
	gtext_thaw(text);
}


void
gdiff_merge_insert_basefile(GdiffMerge *merge)
{
	const GList *dlines_node;
	
	g_return_if_fail(merge != NULL);
	g_return_if_fail(GDIFF_IS_MERGE(merge));

	dlines_node = GDIFF_BASEPANE(merge)->cur_dlines_node;
	if (dlines_node == NULL)
		return;

	insert_base(merge, dlines_node->data);
}

void
gdiff_merge_remove_basefile(GdiffMerge *merge)
{
	const GList *dlines_node;
	
	g_return_if_fail(merge != NULL);
	g_return_if_fail(GDIFF_IS_MERGE(merge));
	
	dlines_node = GDIFF_BASEPANE(merge)->cur_dlines_node;
	if (dlines_node == NULL)
		return;

	remove_base(merge, dlines_node->data);
}

void
gdiff_merge_insert_otherfile(GdiffMerge *merge)
{
	const GList *dlines_node;
	
	g_return_if_fail(merge != NULL);
	g_return_if_fail(GDIFF_IS_MERGE(merge));
	
	dlines_node = GDIFF_BASEPANE(merge)->cur_dlines_node;
	if (dlines_node == NULL)
		return;

	insert_other(merge, dlines_node->data);
}

void
gdiff_merge_remove_otherfile(GdiffMerge *merge)
{
	const GList *dlines_node;
	
	g_return_if_fail(merge != NULL);
	g_return_if_fail(GDIFF_IS_MERGE(merge));
	
	dlines_node = GDIFF_BASEPANE(merge)->cur_dlines_node;
	if (dlines_node == NULL)
		return;

	remove_other(merge, dlines_node->data);
}

void
gdiff_merge_output_file(GdiffMerge *merge, const char *filename)
{
	g_return_if_fail(merge != NULL);
	g_return_if_fail(GDIFF_IS_MERGE(merge));
	g_return_if_fail(filename != NULL && filename[0] != '\0');

	guts_output_file(merge, filename);
}



/** Internal functions **/
/**
 * guts_merge_display:
 * Show the diff result in merge.
 **/
static void
guts_merge_display(GdiffMerge *merge)
{
	draw_text(merge);

	if (PANE_PREF(merge).show_line_num == TRUE) {
		show_hide_numbers(merge, TRUE);
	}
}

/**
 * draw_text:
 * Draw text with coloring different parts.
 **/
static void
draw_text(GdiffMerge *merge)
{
	DiffFiles *dfiles = PANE_DFILES(merge);
	const FileInfo *fi = dfiles_get_fileinfo(dfiles, basefile, TRUE);
	const FileInfo *fi_o = dfiles_get_fileinfo(dfiles, otherfile, TRUE);
	MBuffer *mbuf = PANE_MBUF(merge, basefile);
	const char *buf_pt;
	int buf_ln;
	int buf_lenb;
	DTextMap *dtmap = GDIFF_ONEPANE(merge)->dtmap;
	GtkWidget *text = GDIFF_ONEPANE(merge)->text;
	GdkFont *font = text->style->font;
	GList *node;/* node of DiffLines list */
	FontProp fprop;
	int oldpos = 0;
	int pos;

	if (fi->buf == NULL && fi_o->buf == NULL) /* Maybe, the file has been deleted. */
		return;

 	fprop.font = font;
	mbuf_goto_top(mbuf);
	gtext_freeze(text);
	for (node = dfiles->dlines_list; node; node = node->next) {
		const DiffLines *dlines = node->data;
		/* Use local variables for readability */
		int begin = dlines->between[basefile].begin;
		int end = dlines->between[basefile].end;

		fprop.fg = fprop.bg = NULL;
		buf_ln = mbuf->cur_ln;
		buf_pt = mbuf->cur_pt;
		buf_lenb = mbuf_goto_line(mbuf, begin) - buf_pt;
		pos = gtext_insert_buf(text, &fprop, buf_pt, mbuf->cur_pt - buf_pt);
		dtmap_append_displn(dtmap, DA_COMMON, oldpos, pos - oldpos,
							buf_pt, buf_lenb, begin - buf_ln);
		oldpos = pos;

		buf_pt = mbuf->cur_pt;
		if (end == 0) {/* special */
			g_assert(IS_OTHER_ONLY(basefile, dlines->difftype));
			dtmap_append_displn(dtmap, DA_O_ONLY|DA_O_ONLY_TOP, 0, 0, 0, 0, 0);
			/* mark here (after) */
			fprop.fg = fprop.bg = &GTK_WIDGET(text)->style->bg[GTK_STATE_INSENSITIVE];
			pos = gtext_insert_buf(text, &fprop, marktext, marktext_lenb);
			dtmap_append_displn(dtmap, DA_MARK, oldpos, pos - oldpos,
								marktext, marktext_lenb, 1);
			oldpos = pos;
			continue;
		} else {
			buf_lenb = mbuf_goto_line(mbuf, end + 1) - buf_pt;
		}
		if (IS_OTHER_ONLY(basefile, dlines->difftype)) {
			pos = gtext_insert_buf(text, &fprop, buf_pt, buf_lenb);
			dtmap_append_displn(dtmap, DA_O_ONLY, oldpos, pos - oldpos,
								buf_pt, buf_lenb, end+1 - begin);
			oldpos = pos;

			/* mark here (after) */
			fprop.fg = fprop.bg = &GTK_WIDGET(text)->style->bg[GTK_STATE_INSENSITIVE];
			pos = gtext_insert_buf(text, &fprop, marktext, marktext_lenb);
			dtmap_append_displn(dtmap, DA_MARK, oldpos, pos - oldpos,
								marktext, marktext_lenb, 1);
			oldpos = pos;
		} else {
			DispAttr attr;
			int num_chars;
			
			/* mark here (before) */
			fprop.fg = fprop.bg = &GTK_WIDGET(text)->style->bg[GTK_STATE_INSENSITIVE];
			pos = gtext_insert_buf(text, &fprop, marktext, marktext_lenb);
			dtmap_append_displn(dtmap, DA_MARK, oldpos, pos - oldpos,
								marktext, marktext_lenb, 1);
			oldpos = pos;

			attr = dlines->difftype == CHANGE ? DA_CHANGE : DA_ONLY;
			num_chars = get_num_chars(buf_pt, buf_lenb, !GTK_TEXT(text)->use_wchar);
			dtmap_append_displn(dtmap, attr|DA_HIDE, oldpos, num_chars,
								buf_pt, buf_lenb, end+1 - begin);
		}
	}
	/* Draw the remained part */
	fprop.fg = fprop.bg = NULL;
	buf_ln = mbuf->cur_ln;
	buf_pt = mbuf->cur_pt;
	buf_lenb = mbuf_goto_line(mbuf, fi->nlines + 1) - buf_pt;
	pos = gtext_insert_buf(text, &fprop, buf_pt, buf_lenb);
	dtmap_append_displn(dtmap, DA_COMMON, oldpos, pos - oldpos,
						buf_pt, buf_lenb, fi->nlines+1 - buf_ln);
	gtext_thaw(text);
}


/**
 * show_hide_numbers:
 * Show(or hide) line numbers on the head of each lines.
 * This takes care of two files(merged text) at the same time.
 **/
static void
show_hide_numbers(GdiffMerge *merge, gboolean b_show)
{
	/*XXX TODO*/
}

/**
 * text_click_cb:
 * Signal handler for button_click on text widget.
 * This is connected after default handler,
 * so I can get the clicked position (point in text widget term).
 * Calculate line number from the point,
 * and map it to line number on the buffer.
 **/
static gint
text_click_cb(GtkWidget *text, GdkEventButton *event, gpointer data)
{
	GdiffMerge *merge = data;
	DiffFiles *dfiles = PANE_DFILES(merge);
	DTextMap *dtmap = GDIFF_ONEPANE(merge)->dtmap;

	if (event->button == 1) {
		const GList *dlines_node;
		int index, tln, bln;

		index = gtext_get_cursor_pos(text);
		tln = gtext_line_from_index(text, index);
		bln = dtmap_map_t2b(dtmap, tln);
#ifdef DEBUG
		g_print("tln=%d, bln=%d\n", tln, bln);
#endif
		dlines_node = dfiles_find_includel(dfiles, basefile, bln);
		if (dlines_node == NULL && bln == 1)
			dlines_node = dfiles_find_rel_curl(dfiles, basefile, 0);/* exception */

		if (dlines_node) {
			gtk_signal_emit_by_name(GTK_OBJECT(merge), "select_dlines",
									basefile, bln);
		}
	}

	return FALSE;
}


static void
insert_base(GdiffMerge *merge, const DiffLines *dlines)
{
	MBuffer *mbuf = PANE_MBUF(merge, basefile);
	const char *buf_pt;
	int buf_lenb;
	DTextMap *dtmap = GDIFF_ONEPANE(merge)->dtmap;
	GtkWidget *text = GDIFF_ONEPANE(merge)->text;
	FontProp fprop;
	GdkFont *font = text->style->font;
	GdkColor *fg_color = &PANE_PREF(merge).diff_fg[basefile];
	GdkColor *bg_color = &PANE_PREF(merge).diff_bg[basefile];
	DispLines *displn;
	DispLines *displn_mark;
	int pos, newpos;
	/* Use local variables for readability */
	int begin = dlines->between[basefile].begin;
	int end = dlines->between[basefile].end;

 	fprop.font = font;
	fprop.fg = fg_color;
	fprop.bg = bg_color;

	displn = dtmap_lookup_by_bufln(dtmap, begin);
	g_assert(displn != NULL);
	if (dlines->difftype == CHANGE) {
		g_assert(displn->attr & DA_CHANGE);
		if (displn->attr & DA_HIDE) {
			gtext_freeze(text);
			displn_mark = dtmap_lookup_prev(dtmap, displn);
			if (displn_mark && (displn_mark->attr & DA_MARK)) {
				/* remove mark */
				remove_displn(dtmap, text, displn_mark);
			} else {
				pos = displn->pos;
				gtext_set_point(text, pos);
			}
			/* show my parts */
			buf_pt = mbuf_goto_line(mbuf, begin);
			buf_lenb = mbuf_goto_line(mbuf, end + 1) - buf_pt;
			gtext_insert_buf(text, &fprop, buf_pt, buf_lenb);
			gtext_thaw(text);
			/* hack to show */
			dtmap_show_displn(dtmap, displn);
		} else {
#ifdef DEBUG			
			g_print("already shown");
#endif			
			return;
		}
	} else if (IS_THIS_ONLY(basefile, dlines->difftype)) {
		displn_mark = dtmap_lookup_prev(dtmap, displn);
		if (displn_mark && (displn_mark->attr & DA_MARK)) {
			/* remove mark */
			gtext_freeze(text);
			remove_displn(dtmap, text, displn_mark);
			/* show my parts */
			buf_pt = mbuf_goto_line(mbuf, begin);
			buf_lenb = mbuf_goto_line(mbuf, end + 1) - buf_pt;
			newpos = gtext_insert_buf(text, &fprop, buf_pt, buf_lenb);
			gtext_thaw(text);
			/* hack to show */
			dtmap_show_displn(dtmap, displn);
		} else {
#ifdef DEBUG			
			g_print("already shown");
#endif			
			return;
		}
	} else {
		return;/* do nothing */
	}
}

static void
insert_other(GdiffMerge *merge, const DiffLines *dlines)
{
	MBuffer *mbuf_o = PANE_MBUF(merge, otherfile);
	const char *buf_pt_o;
	int buf_lenb_o;
	DTextMap *dtmap = GDIFF_ONEPANE(merge)->dtmap;
	GtkWidget *text = GDIFF_ONEPANE(merge)->text;
	FontProp fprop;
	GdkFont *font = text->style->font;
	GdkColor *fg_color_o = &PANE_PREF(merge).diff_fg[otherfile];
	GdkColor *bg_color_o = &PANE_PREF(merge).diff_bg[otherfile];
	DispLines *displn;
	DispLines *displn_mark;
	int pos, newpos;
	/* Use local variables for readability */
	int begin = dlines->between[basefile].begin;
	int begin_o = dlines->between[otherfile].begin;
	int end_o = dlines->between[otherfile].end;

	fprop.font = font;
	fprop.fg = fg_color_o;
	fprop.bg = bg_color_o;

	displn = dtmap_lookup_by_bufln(dtmap, begin);
	g_assert(displn != NULL);
	if (dlines->difftype == CHANGE) {
		g_assert(displn->attr & DA_CHANGE);
		displn_mark = dtmap_lookup_prev(dtmap, displn);
		if (displn_mark && (displn_mark->attr & DA_MARK)) {
			gtext_freeze(text);
			/* remove mark */
			pos = displn_mark->pos;
			remove_displn(dtmap, text, displn_mark);
			/* show my parts */
			buf_pt_o = mbuf_goto_line(mbuf_o, begin_o);
			buf_lenb_o = mbuf_goto_line(mbuf_o, end_o + 1) - buf_pt_o;
			newpos = gtext_insert_buf(text, &fprop, buf_pt_o, buf_lenb_o);
			gtext_thaw(text);
			dtmap_insert_displn(dtmap, DA_CHANGE_O, pos, newpos - pos,
								buf_pt_o, buf_lenb_o, end_o+1 - begin_o);
		} else {
			DispLines *prev = dtmap_lookup_prev(dtmap, displn);
			DispLines *next = dtmap_lookup_next(dtmap, displn);

			if (prev && (prev->attr & DA_CHANGE_O)) {
				if (mbuf_goto_line(mbuf_o, begin_o) == mbuf_goto_top(MBUFFER(prev))) {
#ifdef DEBUG					
					g_print("already shown");
#endif					
					return;
				}
			} else if (next && (next->attr & DA_CHANGE_O)) {
				if (mbuf_goto_line(mbuf_o, begin_o) == mbuf_goto_top(MBUFFER(next))) {
#ifdef DEBUG					
					g_print("already shown");
#endif					
					return;
				}
			} else {
				gtext_freeze(text);
				pos = displn->pos + displn->len;
				gtext_set_point(text, pos);
				/* show my parts (append) */
				buf_pt_o = mbuf_goto_line(mbuf_o, begin_o);
				buf_lenb_o = mbuf_goto_line(mbuf_o, end_o + 1) - buf_pt_o;
				newpos = gtext_insert_buf(text, &fprop, buf_pt_o, buf_lenb_o);
				gtext_thaw(text);
				dtmap_insert_displn(dtmap, DA_CHANGE_O, pos, newpos - pos,
									buf_pt_o, buf_lenb_o, end_o+1 - begin_o);
			}
		}
	} else if (IS_OTHER_ONLY(basefile, dlines->difftype)) {
		g_assert(displn->attr & DA_O_ONLY);
		displn_mark = dtmap_lookup_next(dtmap, displn);
		if (displn_mark && (displn_mark->attr & DA_MARK)) {
			/* remove mark */
			gtext_freeze(text);
			pos = displn_mark->pos;
			remove_displn(dtmap, text, displn_mark);
			/* show my parts */
			buf_pt_o = mbuf_goto_line(mbuf_o, begin_o);
			buf_lenb_o = mbuf_goto_line(mbuf_o, end_o + 1) - buf_pt_o;
			newpos = gtext_insert_buf(text, &fprop, buf_pt_o, buf_lenb_o);
			gtext_thaw(text);
			dtmap_insert_displn(dtmap, DA_ONLY_O, pos, newpos - pos,
								buf_pt_o, buf_lenb_o, end_o+1 - begin_o);
		} else {
#ifdef DEBUG			
			g_print("already shown");
#endif			
			return;
		}
	} else {
		return;/* do nothing */
	}
}

static void
remove_base(GdiffMerge *merge, const DiffLines *dlines)
{
	MBuffer *mbuf_o = PANE_MBUF(merge, otherfile);
	DTextMap *dtmap = GDIFF_ONEPANE(merge)->dtmap;
	GtkWidget *text = GDIFF_ONEPANE(merge)->text;
	FontProp fprop;
	GdkFont *font = text->style->font;
	DispLines *displn;
	int pos, newpos;
	/* Use local variables for readability */
	int begin = dlines->between[basefile].begin;
	int begin_o = dlines->between[otherfile].begin;

	fprop.font = font;
	fprop.fg = fprop.bg = &GTK_WIDGET(text)->style->bg[GTK_STATE_INSENSITIVE];
	
	displn = dtmap_lookup_by_bufln(dtmap, begin);
	g_assert(displn != NULL);
	if (displn->attr & DA_HIDE) {
#ifdef DEBUG		
		g_print("already hide");
#endif		
		return;
	}
	if (IS_THIS_ONLY(basefile, dlines->difftype)) {
		/* insert mark */
		pos = displn->pos;
		gtext_freeze(text);
		gtext_set_point(text, pos);
		newpos = gtext_insert_buf(text, &fprop, marktext, marktext_lenb);
		dtmap_insert_displn(dtmap, DA_MARK, pos, newpos - pos,
							marktext, marktext_lenb, 1);
		/* hide my parts */
		gtext_forward_delete(text, displn->len);
		gtext_thaw(text);
		/* hack to hide */
		dtmap_hide_displn(dtmap, displn);
	} else if (dlines->difftype == CHANGE) {
		DispLines *prev = dtmap_lookup_prev(dtmap, displn);
		DispLines *next = dtmap_lookup_next(dtmap, displn);

		g_assert(displn->attr & DA_CHANGE);
		gtext_freeze(text);
		if (prev && (prev->attr & DA_CHANGE_O)) {
			if (mbuf_goto_line(mbuf_o, begin_o) == mbuf_goto_top(MBUFFER(prev))) {
				/* hide my parts */
				pos = displn->pos;
				gtext_set_point(text, pos);
				gtext_forward_delete(text, displn->len);
				/* hack to hide */
				dtmap_hide_displn(dtmap, displn);
			} else {
				g_assert_not_reached();
			}
		} else if (next && (next->attr & DA_CHANGE_O)) {
			if (mbuf_goto_line(mbuf_o, begin_o) == mbuf_goto_top(MBUFFER(next))) {
				/* hide my parts */
				pos = displn->pos;
				gtext_set_point(text, pos);
				gtext_forward_delete(text, displn->len);
				/* hack to hide */
				dtmap_hide_displn(dtmap, displn);
			} else {
				g_assert_not_reached();
			}
		} else {
			/* insert mark */
			pos = displn->pos;
			gtext_set_point(text, pos);
			newpos = gtext_insert_buf(text, &fprop, marktext, marktext_lenb);
			dtmap_insert_displn(dtmap, DA_MARK, pos, newpos - pos,
								marktext, marktext_lenb, 1);
			/* hide my parts */
			gtext_forward_delete(text, displn->len);
			/* hack to hide */
			dtmap_hide_displn(dtmap, displn);
		}
		gtext_thaw(text);
	} else {
		return;/* do nothing */
	}
}

static void
remove_other(GdiffMerge *merge, const DiffLines *dlines)
{
	MBuffer *mbuf_o = PANE_MBUF(merge, otherfile);
	DTextMap *dtmap = GDIFF_ONEPANE(merge)->dtmap;
	GtkWidget *text = GDIFF_ONEPANE(merge)->text;
	FontProp fprop;
	GdkFont *font = text->style->font;
	DispLines *displn;
	DispLines *displn_o;
	int pos, newpos;
	/* Use local variables for readability */
	int begin = dlines->between[basefile].begin;
	int begin_o = dlines->between[otherfile].begin;

	fprop.font = font;
	fprop.fg = fprop.bg = &GTK_WIDGET(text)->style->bg[GTK_STATE_INSENSITIVE];

	displn = dtmap_lookup_by_bufln(dtmap, begin);
	g_assert(displn != NULL);
	if (dlines->difftype == CHANGE) {
		DispLines *prev = dtmap_lookup_prev(dtmap, displn);
		DispLines *next = dtmap_lookup_next(dtmap, displn);

		g_assert(displn->attr & DA_CHANGE);
		if (displn->attr & DA_HIDE) {
			if (prev && (prev->attr & DA_CHANGE_O)) {
				if (mbuf_goto_line(mbuf_o, begin_o) == mbuf_goto_top(MBUFFER(prev))) {
					/* hide my parts */
					gtext_freeze(text);
					remove_displn(dtmap, text, prev);
				}
			} else if (next && (next->attr & DA_CHANGE_O)) {
				if (mbuf_goto_line(mbuf_o, begin_o) == mbuf_goto_top(MBUFFER(next))) {
					/* hide my parts */
					gtext_freeze(text);
					remove_displn(dtmap, text, next);
				}
			} else {
#ifdef DEBUG				
				g_print("already hidden");
#endif				
				return;
			}
			/* insert mark */
			pos = displn->pos;
			gtext_set_point(text, pos);
			newpos = gtext_insert_buf(text, &fprop, marktext, marktext_lenb);
			dtmap_insert_displn(dtmap, DA_MARK, pos, newpos - pos,
								marktext, marktext_lenb, 1);
			gtext_thaw(text);
		} else {
			if (prev && (prev->attr & DA_CHANGE_O)) {
				if (mbuf_goto_line(mbuf_o, begin_o) == mbuf_goto_top(MBUFFER(prev))) {
					/* hide my parts */
					gtext_freeze(text);
					remove_displn(dtmap, text, prev);
					gtext_thaw(text);
				}
			} else if (next && (next->attr & DA_CHANGE_O)) {
				if (mbuf_goto_line(mbuf_o, begin_o) == mbuf_goto_top(MBUFFER(next))) {
					/* hide my parts */
					gtext_freeze(text);
					remove_displn(dtmap, text, next);
					gtext_thaw(text);
				}
			} else {
#ifdef DEBUG				
				g_print("already hidden");
#endif				
				return;
			}
		}
	} else if (IS_OTHER_ONLY(basefile, dlines->difftype)) {
		g_assert(displn->attr & DA_O_ONLY);
		displn_o = dtmap_lookup_next(dtmap, displn);
		if (displn_o && (displn_o->attr & DA_ONLY_O)) {
			/* hide my parts */
			gtext_freeze(text);
			pos = displn_o->pos;
			remove_displn(dtmap, text, displn_o);
			/* insert mark */
			newpos = gtext_insert_buf(text, &fprop, marktext, marktext_lenb);
			dtmap_insert_displn(dtmap, DA_MARK, pos, newpos - pos,
								marktext, marktext_lenb, 1);
			gtext_thaw(text);
		} else if (displn_o && (displn_o->attr & DA_MARK)) {
#ifdef DEBUG			
			g_print("already hidden");
#endif			
			return;
		} else {
			return;/* do nothing */
		}
	} else {
		return;/* do nothing */
	}
}

static void
remove_displn(DTextMap *dtmap, GtkWidget *text, DispLines *displn)
{
	int pos;
	
	pos = displn->pos;
	gtext_set_point(text, pos);
	gtext_forward_delete(text, displn->len);
	dtmap_remove_displn(dtmap, displn);
}

static void
guts_output_file(GdiffMerge *merge, const char *filename)
{
	DTextMap *dtmap = GDIFF_ONEPANE(merge)->dtmap;
	DispLines *displn;
	FILE *fp = NULL;

	fp = fopen(filename, "w");
	if (fp == NULL) {
		g_message("can't write to %s\n", filename);
		return;
	}

	for (displn = dtmap_first_displn(dtmap, DA_ANY); displn; displn = dtmap_next_displn(dtmap, DA_ANY)) {
		const char *buf_pt;
		
		if (displn->attr & DA_HIDE)
			continue;
		if (displn->attr & DA_MARK)
			continue;
		
		buf_pt = mbuf_goto_top(MBUFFER(displn));
		fwrite(buf_pt, mbuf_goto_bottom(MBUFFER(displn)) - buf_pt, 1, fp);
	}
	if (fp)
		fclose(fp);
}
