/*
 * Merge view widget module
 *
 * Copyright INOUE Seiichiro <inoue@ainet.or.jp>, licensed under the GPL.
 */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <gnome.h>
#include "diff.h"
#include "gui.h"
#include "mergeview.h"
#include "merge-widget.h"
#include "dtextmap.h"
#include "merge-menu.h"
#include "guimisc.h"


/* Constant number */
enum {
	ARG_0,
	ARG_BASEPANE,
	ARG_AUXPANE
};

/* Private function declarations */
static void gdiff_mergeview_class_init(GdiffMergeViewClass *klass);
static void gdiff_mergeview_set_arg(GtkObject *object, GtkArg *arg, guint arg_id);
static void gdiff_mergeview_get_arg(GtkObject *object, GtkArg *arg, guint arg_id);
static void gdiff_mergeview_init(GdiffMergeView *mergeview);

static void gdiff_mergeview_set_basepane(GdiffMergeView *mergeview, GdiffBasePane *basepane);
static void gdiff_mergeview_set_auxpane(GdiffMergeView *mergeview, GdiffBasePane *auxpane);

static void create_panes(GdiffMergeView *mergeview, DiffDir *diffdir, DiffFiles *dfiles);
static void create_overview(GdiffMergeView *mergeview, DiffFiles *dfiles);
static void create_rmenu(GdiffMergeView *mergeview);

static void merge_move_cb(GdiffMerge *merge, MoveDiff mv_diff, gpointer data);
static void twopane_move_cb(GdiffTwoPane *twopane, MoveDiff mv_diff, gpointer data);
static void merge_select_dl_cb(GdiffMerge *merge, WhichFile whichfile, int ln, gpointer data);
static void twopane_select_dl_cb(GdiffTwoPane *twopane, WhichFile whichfile, int ln, gpointer data);

static void update_statusbar(GdiffMergeView *mergeview);


static GtkHBoxClass *parent_class = NULL;

GtkType
gdiff_mergeview_get_type(void)
{
  static GtkType mergeview_type = 0;

  if (!mergeview_type) {
      static const GtkTypeInfo mergeview_info = {
		  "GdiffMergeView",
		  sizeof(GdiffMergeView),
		  sizeof(GdiffMergeViewClass),
		  (GtkClassInitFunc)gdiff_mergeview_class_init,
		  (GtkObjectInitFunc)gdiff_mergeview_init,
		  /* reserved_1 */ NULL,
		  /* reserved_2 */ NULL,
		  (GtkClassInitFunc)NULL,
      };
      mergeview_type = gtk_type_unique(GTK_TYPE_HBOX, &mergeview_info);
  }
  
  return mergeview_type;
}

static void
gdiff_mergeview_class_init(GdiffMergeViewClass *klass)
{
	GtkObjectClass *object_class;
	GtkWidgetClass *widget_class;

	gtk_object_add_arg_type("GdiffMergeView::basepane",
							GDIFF_TYPE_MERGE,
							GTK_ARG_READWRITE,
							ARG_BASEPANE);
	gtk_object_add_arg_type("GdiffMergeView::auxpane",
							GDIFF_TYPE_TWOPANE,
							GTK_ARG_READWRITE,
							ARG_AUXPANE);
	object_class = (GtkObjectClass*)klass;
	widget_class = (GtkWidgetClass*)klass;

	parent_class = gtk_type_class(GTK_TYPE_HBOX);

	object_class->set_arg = gdiff_mergeview_set_arg;
	object_class->get_arg = gdiff_mergeview_get_arg;
}

static void
gdiff_mergeview_set_arg(GtkObject *object, GtkArg *arg, guint arg_id)
{
	GdiffMergeView *mergeview;
  
	mergeview = GDIFF_MERGEVIEW(object);
  
	switch (arg_id) {
	case ARG_BASEPANE:
		gdiff_mergeview_set_basepane(mergeview, GTK_VALUE_POINTER(*arg));
		break;
	case ARG_AUXPANE:
		gdiff_mergeview_set_auxpane(mergeview, GTK_VALUE_POINTER(*arg));
		break;
	default:
		break;
	}
}

static void
gdiff_mergeview_get_arg(GtkObject *object, GtkArg *arg, guint arg_id)
{
	GdiffMergeView *mergeview;
  
	mergeview = GDIFF_MERGEVIEW(object);
  
	switch (arg_id) {
    case ARG_BASEPANE:
		GTK_VALUE_POINTER(*arg) = mergeview->merge;
		break;
    case ARG_AUXPANE:
		GTK_VALUE_POINTER(*arg) = mergeview->twopane;
		break;
    default:
		arg->type = GTK_TYPE_INVALID;
		break;
    }
}

static void
gdiff_mergeview_init(GdiffMergeView *mergeview)
{
	mergeview->vpaned = NULL;
	mergeview->merge = NULL;
	mergeview->twopane = NULL;
	mergeview->overview = NULL;

	mergeview->gdwin = NULL;

	mergeview->is_under_dir = FALSE;
}


GtkWidget*
gdiff_mergeview_new(DiffDir *diffdir, DiffFiles *dfiles, gboolean is_under_dir)
{
	GdiffMergeView *mergeview;

	mergeview = gtk_type_new(GDIFF_TYPE_MERGEVIEW);

	create_panes(mergeview, diffdir, dfiles);
	create_overview(mergeview, dfiles);
	create_rmenu(mergeview);

	mergeview->is_under_dir = is_under_dir;
	
	return GTK_WIDGET(mergeview);
}



/** Internal functions **/
static void
gdiff_mergeview_set_basepane(GdiffMergeView *mergeview, GdiffBasePane *basepane)
{
	g_return_if_fail(mergeview != NULL);
	g_return_if_fail(GDIFF_IS_MERGEVIEW(mergeview));
	g_return_if_fail(mergeview->merge == NULL);
	
	if (basepane)
		g_return_if_fail(GDIFF_IS_MERGE(basepane));

	mergeview->merge = basepane;
}

static void
gdiff_mergeview_set_auxpane(GdiffMergeView *mergeview, GdiffBasePane *auxpane)
{
	g_return_if_fail(mergeview != NULL);
	g_return_if_fail(GDIFF_IS_MERGEVIEW(mergeview));
	g_return_if_fail(mergeview->twopane == NULL);
	
	if (auxpane)
		g_return_if_fail(GDIFF_IS_TWOPANE(auxpane));

	mergeview->twopane = GDIFF_TWOPANE(auxpane);
}

static void
create_panes(GdiffMergeView *mergeview, DiffDir *diffdir, DiffFiles *dfiles)
{
	GtkWidget *vpaned;
	GtkWidget *merge;
	GtkWidget *twopane;
	
	vpaned = gtk_vpaned_new();
	mergeview->vpaned = GTK_VPANED(vpaned);
	gtk_box_pack_start(GTK_BOX(mergeview), vpaned, TRUE, TRUE, 0);
	gtk_widget_show(vpaned);
	
	merge = gdiff_merge_new(diffdir, dfiles);
	gdiff_mergeview_set_basepane(mergeview, GDIFF_BASEPANE(merge));
	gtk_paned_pack1(GTK_PANED(vpaned), merge, TRUE, TRUE);
	mergeview->merge_move_handler
		= gtk_signal_connect(GTK_OBJECT(merge), "move_diff",
							 GTK_SIGNAL_FUNC(merge_move_cb), mergeview);
	mergeview->merge_select_dl_handler
		= gtk_signal_connect(GTK_OBJECT(merge), "select_dlines",
							 GTK_SIGNAL_FUNC(merge_select_dl_cb), mergeview);
	gtk_widget_show(merge);

	twopane = gdiff_twopane_new(diffdir, dfiles);
	mergeview->twopane = GDIFF_TWOPANE(twopane);
	gtk_paned_pack2(GTK_PANED(vpaned), twopane, TRUE, TRUE);
	mergeview->twop_move_handler
		= gtk_signal_connect(GTK_OBJECT(twopane), "move_diff",
							 GTK_SIGNAL_FUNC(twopane_move_cb), mergeview);
	mergeview->twop_select_dl_handler
		= gtk_signal_connect(GTK_OBJECT(twopane), "select_dlines",
							 GTK_SIGNAL_FUNC(twopane_select_dl_cb), mergeview);
	gtk_widget_show(twopane);
}


/**
 * create_overview:
 * Create GdiffOverview widget.
 * See "gdiffoverview.[ch]" about GdiffOverview widget.
 **/
static void
create_overview(GdiffMergeView *mergeview, DiffFiles *dfiles)
{
	GtkWidget *overview;
	GdiffOnePane *merge = GDIFF_ONEPANE(mergeview->merge);
	DTextMap *dtmap = merge->dtmap;

	overview = gdiff_overview_new(GTK_TEXT(merge->text)->vadj,
								  GTK_TEXT(merge->text)->vadj);
	mergeview->overview = GDIFF_OVERVIEW(overview);
	gdiff_overview_set_color1(GDIFF_OVERVIEW(overview),
							  &PANE_PREF(merge).diff_bg[FIRST_FILE],
							  NULL);
	gdiff_overview_set_color2(GDIFF_OVERVIEW(overview),
							  &PANE_PREF(merge).diff_bg[SECOND_FILE],
							  NULL);

	/* initialize overview widget */
	if (dtmap->total_nl != 0) {
		GList *node;/* node of DiffLines list */

		for (node = dfiles->dlines_list; node; node = node->next) {
			const DiffLines *dlines = node->data;
			/* Use local variables for ease to read */
			int begin1 = dlines->between[FIRST_FILE].begin;
			int end1 = dlines->between[FIRST_FILE].end;
			int begin2 = dlines->between[SECOND_FILE].begin;
			int end2 = dlines->between[SECOND_FILE].end;
			/* Map to text */
			int tbegin1 = dtmap_map_b2t(dtmap, begin1);
			int tend1 = dtmap_map_b2t(dtmap, end1);
			int tbegin2 = dtmap_map_b2t(dtmap, begin2);
			int tend2 = dtmap_map_b2t(dtmap, end2);
			
			gdiff_overview_insert_paintrange(GDIFF_OVERVIEW(overview),
											 (gdouble)tbegin1/dtmap->total_nl,
											 (gdouble)tend1/dtmap->total_nl,
											 (gdouble)tbegin2/dtmap->total_nl,
											 (gdouble)tend2/dtmap->total_nl);
		}
	}
	
	gtk_box_pack_start(GTK_BOX(mergeview), overview, FALSE, FALSE, 0);
	gtk_widget_show(overview);
}

/* right-click menu */
static void
create_rmenu(GdiffMergeView *mergeview)
{
	GdiffOnePane *merge = GDIFF_ONEPANE(mergeview->merge);
	GtkWidget *rmenu;

	rmenu = merge_rmenu_create();
	gnome_popup_menu_attach(rmenu, merge->text, mergeview);
}

static void
merge_move_cb(GdiffMerge *merge, MoveDiff mv_diff, gpointer data)
{
	GdiffMergeView *mergeview = data;
	GdiffTwoPane *twopane = mergeview->twopane;

	/* sync twopane-ext widget */
	gtk_signal_handler_block(GTK_OBJECT(twopane),
							 mergeview->twop_move_handler);
	gdiff_basepane_move_diff(GDIFF_BASEPANE(twopane), mv_diff);
	gtk_signal_handler_unblock(GTK_OBJECT(twopane),
							   mergeview->twop_move_handler);

	update_statusbar(mergeview);
}

static void
twopane_move_cb(GdiffTwoPane *twopane, MoveDiff mv_diff, gpointer data)
{
	GdiffMergeView *mergeview = data;
	GdiffMerge *merge = GDIFF_MERGE(mergeview->merge);

	/* sync merge widget */
	gtk_signal_handler_block(GTK_OBJECT(merge),
							 mergeview->merge_move_handler);
	gdiff_basepane_move_diff(GDIFF_BASEPANE(merge), mv_diff);
	gtk_signal_handler_unblock(GTK_OBJECT(merge),
							   mergeview->merge_move_handler);
}

static void
merge_select_dl_cb(GdiffMerge *merge, WhichFile whichfile, int ln, gpointer data)
{
	GdiffMergeView *mergeview = data;
	GdiffTwoPane *twopane = mergeview->twopane;

	/* sync twopane-ext widget */
	gtk_signal_handler_block(GTK_OBJECT(twopane),
							 mergeview->twop_select_dl_handler);
	gdiff_basepane_select_dlines(GDIFF_BASEPANE(twopane), whichfile, ln);
	gtk_signal_handler_unblock(GTK_OBJECT(twopane),
							   mergeview->twop_select_dl_handler);

	update_statusbar(mergeview);
}

static void
twopane_select_dl_cb(GdiffTwoPane *twopane, WhichFile whichfile, int ln, gpointer data)
{
	GdiffMergeView *mergeview = data;
	GdiffMerge *merge = GDIFF_MERGE(mergeview->merge);

	/* sync merge widget */
	gtk_signal_handler_block(GTK_OBJECT(merge),
							 mergeview->merge_select_dl_handler);
	gdiff_basepane_select_dlines(GDIFF_BASEPANE(merge), whichfile, ln);
	gtk_signal_handler_unblock(GTK_OBJECT(merge),
							   mergeview->merge_select_dl_handler);
}


static void
update_statusbar(GdiffMergeView *mergeview)
{
	GdiffBasePane *merge = GDIFF_BASEPANE(mergeview->merge);
	char *sbar_msg = NULL;/* status-bar message */

	if (merge->cur_dlines_node) {
		const DiffLines *dlines;
		
		dlines = merge->cur_dlines_node->data;
		sbar_msg = make_statusbar_msg(
			GDIFF_MERGEVIEW_FILENAME(mergeview, FIRST_FILE),
			GDIFF_MERGEVIEW_FILENAME(mergeview, SECOND_FILE),
			dlines->between[FIRST_FILE].begin,
			dlines->between[FIRST_FILE].end,
			dlines->between[SECOND_FILE].begin,
			dlines->between[SECOND_FILE].end);
	}
	if (sbar_msg) {
		statusbar_update(mergeview->gdwin, sbar_msg);
		g_free(sbar_msg);
	}
}
