/*
 * One-pane (file)compare view widget module
 *
 * Copyright INOUE Seiichiro <inoue@ainet.or.jp>, licensed under the GPL.
 */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <gnome.h>
#include "diff.h"
#include "gui.h"
#include "onepaneview.h"
#include "onepane-widget.h"
#include "dtextmap.h"
#include "rmenu.h"
#include "guimisc.h"


/* Constant number */
enum {
	ARG_0,
	ARG_BASEPANE
};

/* Private function declarations */
static void gdiff_onepview_class_init(GdiffOnepViewClass *klass);
static void gdiff_onepview_set_arg(GtkObject *object, GtkArg *arg, guint arg_id);
static void gdiff_onepview_get_arg(GtkObject *object, GtkArg *arg, guint arg_id);
static void gdiff_onepview_init(GdiffOnepView *onepview);

static void gdiff_onepview_set_basepane(GdiffOnepView *onepview, GdiffBasePane *basepane);

static void create_onepane(GdiffOnepView *onepview, DiffDir *diffdir, DiffFiles *dfiles);
static void create_overview(GdiffOnepView *onepview, DiffFiles *dfiles);
static void create_rmenu(GdiffOnepView *onepview);

static void move_diff_cb(GdiffOnePane *onepane, MoveDiff mv_diff, gpointer data);
static void update_statusbar(GdiffOnepView *onepview);


static GtkHBoxClass *parent_class = NULL;

GtkType
gdiff_onepview_get_type(void)
{
  static GtkType onepview_type = 0;

  if (!onepview_type) {
      static const GtkTypeInfo onepview_info = {
		  "GdiffOnepView",
		  sizeof(GdiffOnepView),
		  sizeof(GdiffOnepViewClass),
		  (GtkClassInitFunc)gdiff_onepview_class_init,
		  (GtkObjectInitFunc)gdiff_onepview_init,
		  /* reserved_1 */ NULL,
		  /* reserved_2 */ NULL,
		  (GtkClassInitFunc)NULL,
      };
      onepview_type = gtk_type_unique(GTK_TYPE_HBOX, &onepview_info);
  }
  
  return onepview_type;
}

static void
gdiff_onepview_class_init(GdiffOnepViewClass *klass)
{
	GtkObjectClass *object_class;
	GtkWidgetClass *widget_class;

	gtk_object_add_arg_type("GdiffOnepView::basepane",
							GDIFF_TYPE_ONEPANE,
							GTK_ARG_READWRITE,
							ARG_BASEPANE);
	object_class = (GtkObjectClass*)klass;
	widget_class = (GtkWidgetClass*)klass;

	parent_class = gtk_type_class(GTK_TYPE_HBOX);

	object_class->set_arg = gdiff_onepview_set_arg;
	object_class->get_arg = gdiff_onepview_get_arg;
}

static void
gdiff_onepview_set_arg(GtkObject *object, GtkArg *arg, guint arg_id)
{
	GdiffOnepView *onepview;
  
	onepview = GDIFF_ONEPVIEW(object);
  
	switch (arg_id) {
	case ARG_BASEPANE:
		gdiff_onepview_set_basepane(onepview, GTK_VALUE_POINTER(*arg));
		break;
	default:
		break;
	}
}

static void
gdiff_onepview_get_arg(GtkObject *object, GtkArg *arg, guint arg_id)
{
	GdiffOnepView *onepview;
  
	onepview = GDIFF_ONEPVIEW(object);
  
	switch (arg_id) {
    case ARG_BASEPANE:
		GTK_VALUE_POINTER(*arg) = onepview->onepane;
		break;
    default:
		arg->type = GTK_TYPE_INVALID;
		break;
    }
}

static void
gdiff_onepview_init(GdiffOnepView *onepview)
{
	onepview->onepane = NULL;
	onepview->overview = NULL;

	onepview->gdwin = NULL;

	onepview->is_under_dir = FALSE;
}


GtkWidget*
gdiff_onepview_new(DiffDir *diffdir, DiffFiles *dfiles, gboolean is_under_dir)
{
	GdiffOnepView *onepview;

	onepview = gtk_type_new(GDIFF_TYPE_ONEPVIEW);

	create_onepane(onepview, diffdir, dfiles);
	create_overview(onepview, dfiles);
	create_rmenu(onepview);

	onepview->is_under_dir = is_under_dir;
	
	return GTK_WIDGET(onepview);
}



/** Internal functions **/
static void
gdiff_onepview_set_basepane(GdiffOnepView *onepview, GdiffBasePane *basepane)
{
	g_return_if_fail(onepview != NULL);
	g_return_if_fail(GDIFF_IS_ONEPVIEW(onepview));
	g_return_if_fail(onepview->onepane == NULL);
	
	if (basepane)
		g_return_if_fail(GDIFF_IS_ONEPANE(basepane));

	onepview->onepane = basepane;
}

static void
create_onepane(GdiffOnepView *onepview, DiffDir *diffdir, DiffFiles *dfiles)
{
	GtkWidget *onepane;

	onepane = gdiff_onepane_new(diffdir, dfiles);
	gdiff_onepview_set_basepane(onepview, GDIFF_BASEPANE(onepane));

	gtk_box_pack_start(GTK_BOX(onepview), onepane, TRUE, TRUE, 0);
	gtk_signal_connect(GTK_OBJECT(onepane), "move_diff",
					   GTK_SIGNAL_FUNC(move_diff_cb), onepview);
	gtk_widget_show(onepane);
}

/**
 * create_overview:
 * Create GdiffOverview widget.
 * See "gdiffoverview.[ch]" about GdiffOverview widget.
 **/
static void
create_overview(GdiffOnepView *onepview, DiffFiles *dfiles)
{
	GtkWidget *overview;
	GdiffOnePane *onepane = GDIFF_ONEPANE(onepview->onepane);
	DTextMap *dtmap = onepane->dtmap;

	overview = gdiff_overview_new(GTK_TEXT(onepane->text)->vadj,
								  GTK_TEXT(onepane->text)->vadj);
	onepview->overview = GDIFF_OVERVIEW(overview);
	gdiff_overview_set_color1(GDIFF_OVERVIEW(overview),
							  &PANE_PREF(onepane).diff_bg[FIRST_FILE],
							  NULL);
	gdiff_overview_set_color2(GDIFF_OVERVIEW(overview),
							  &PANE_PREF(onepane).diff_bg[SECOND_FILE],
							  NULL);

	/* initialize overview widget */
	if (dtmap->total_nl != 0) {
		GList *node;/* node of DiffLines list */

		for (node = dfiles->dlines_list; node; node = node->next) {
			const DiffLines *dlines = node->data;
			/* Use local variables for ease to read */
			int begin1 = dlines->between[FIRST_FILE].begin;
			int end1 = dlines->between[FIRST_FILE].end;
			int begin2 = dlines->between[SECOND_FILE].begin;
			int end2 = dlines->between[SECOND_FILE].end;
			/* Map to text */
			int tbegin1 = dtmap_map_b2t(dtmap, begin1);
			int tend1 = dtmap_map_b2t(dtmap, end1);
			int tbegin2 = dtmap_map_b2t(dtmap, begin2);
			int tend2 = dtmap_map_b2t(dtmap, end2);
			
			gdiff_overview_insert_paintrange(GDIFF_OVERVIEW(overview),
											 (gdouble)tbegin1/dtmap->total_nl,
											 (gdouble)tend1/dtmap->total_nl,
											 (gdouble)tbegin2/dtmap->total_nl,
											 (gdouble)tend2/dtmap->total_nl);
		}
	}
	
	gtk_box_pack_start(GTK_BOX(onepview), overview, FALSE, FALSE, 0);
	gtk_widget_show(overview);
}

/* right-click menu */
static void
create_rmenu(GdiffOnepView *onepview)
{
	GdiffOnePane *onepane = GDIFF_ONEPANE(onepview->onepane);
	GtkWidget *rmenu;

	/* second argument doesn't matter */
	rmenu = rmenu_create(GTK_WIDGET(onepane), GINT_TO_POINTER(0));
	gnome_popup_menu_attach(rmenu, onepane->text, onepview);
}

/**
 * move_diff_cb:
 * Callback for "move_diff" signal of GdiffOnePane widget.
 * "move_diff" itself is handled by GdiffOnePane widget.
 * GdiffOnepView widget's responsibility is to take care of statusbar.
 **/
static void
move_diff_cb(GdiffOnePane *onepane, MoveDiff mv_diff, gpointer data)
{
	GdiffOnepView *onepview = data;

	update_statusbar(onepview);
}

static void
update_statusbar(GdiffOnepView *onepview)
{
	GdiffBasePane *onepane = GDIFF_BASEPANE(onepview->onepane);
	char *sbar_msg = NULL;/* status-bar message */

	if (onepane->cur_dlines_node) {
		const DiffLines *dlines;
		
		dlines = onepane->cur_dlines_node->data;
		sbar_msg = make_statusbar_msg(
			GDIFF_ONEPVIEW_FILENAME(onepview, FIRST_FILE),
			GDIFF_ONEPVIEW_FILENAME(onepview, SECOND_FILE),
			dlines->between[FIRST_FILE].begin,
			dlines->between[FIRST_FILE].end,
			dlines->between[SECOND_FILE].begin,
			dlines->between[SECOND_FILE].end);
	}
	if (sbar_msg) {
		statusbar_update(onepview->gdwin, sbar_msg);
		g_free(sbar_msg);
	}
}
