/*
 * Two-pane (file)compare view widget module.
 *
 * Copyright INOUE Seiichiro <inoue@ainet.or.jp>, licensed under the GPL.
 */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <gnome.h>
#include "diff.h"
#include "gui.h"
#include "twopaneview.h"
#include "twopane-widget.h"
#include "rmenu.h"
#include "guimisc.h"


/* Constant number */
enum {
	ARG_0,
	ARG_BASEPANE
};

/* Private function declarations */
static void gdiff_twopview_class_init(GdiffTwopViewClass *klass);
static void gdiff_twopview_set_arg(GtkObject *object, GtkArg *arg, guint arg_id);
static void gdiff_twopview_get_arg(GtkObject *object, GtkArg *arg, guint arg_id);
static void gdiff_twopview_init(GdiffTwopView *twopview);
static void gdiff_twopview_destroy(GtkObject *object);

static void gdiff_twopview_set_basepane(GdiffTwopView *twopview, GdiffBasePane *basepane);

static void create_twopane(GdiffTwopView *twopview, DiffDir *diffdir, DiffFiles *dfiles);
static void create_overview(GdiffTwopView *twopview, DiffFiles *dfiles);
static void create_vscrollbar(GdiffTwopView *twopview, DiffFiles *dfiles);
static void vscrollbar_vchanged(GtkAdjustment *vadj, gpointer data);
static void create_rmenu(GdiffTwopView *twopview);

static void move_diff_cb(GdiffTwoPane *twopane, MoveDiff mv_diff, gpointer data);
static void update_statusbar(GdiffTwopView *twopview);


static GtkHBoxClass *parent_class = NULL;

GtkType
gdiff_twopview_get_type(void)
{
  static GtkType twopview_type = 0;

  if (!twopview_type) {
      static const GtkTypeInfo twopview_info = {
		  "GdiffTwopView",
		  sizeof(GdiffTwopView),
		  sizeof(GdiffTwopViewClass),
		  (GtkClassInitFunc)gdiff_twopview_class_init,
		  (GtkObjectInitFunc)gdiff_twopview_init,
		  /* reserved_1 */ NULL,
		  /* reserved_2 */ NULL,
		  (GtkClassInitFunc)NULL,
      };
      twopview_type = gtk_type_unique(GTK_TYPE_HBOX, &twopview_info);
  }
  
  return twopview_type;
}

static void
gdiff_twopview_class_init(GdiffTwopViewClass *klass)
{
	GtkObjectClass *object_class;
	GtkWidgetClass *widget_class;

	gtk_object_add_arg_type("GdiffTwopView::basepane",
							GDIFF_TYPE_TWOPANE,
							GTK_ARG_READWRITE,
							ARG_BASEPANE);
	object_class = (GtkObjectClass*)klass;
	widget_class = (GtkWidgetClass*)klass;

	parent_class = gtk_type_class(GTK_TYPE_HBOX);

	object_class->set_arg = gdiff_twopview_set_arg;
	object_class->get_arg = gdiff_twopview_get_arg;
	object_class->destroy = gdiff_twopview_destroy;
}

static void
gdiff_twopview_set_arg(GtkObject *object, GtkArg *arg, guint arg_id)
{
	GdiffTwopView *twopview;
  
	twopview = GDIFF_TWOPVIEW(object);
  
	switch (arg_id) {
	case ARG_BASEPANE:
		gdiff_twopview_set_basepane(twopview, GTK_VALUE_POINTER(*arg));
		break;
	default:
		break;
	}
}

static void
gdiff_twopview_get_arg(GtkObject *object, GtkArg *arg, guint arg_id)
{
	GdiffTwopView *twopview;
  
	twopview = GDIFF_TWOPVIEW(object);
  
	switch (arg_id) {
    case ARG_BASEPANE:
		GTK_VALUE_POINTER(*arg) = twopview->twopane;
		break;
    default:
		arg->type = GTK_TYPE_INVALID;
		break;
    }
}

static void
gdiff_twopview_init(GdiffTwopView *twopview)
{
	twopview->twopane = NULL;
	twopview->overview = NULL;
	twopview->vscrollboth = NULL;

	twopview->gdwin = NULL;

	twopview->is_under_dir = FALSE;
}


GtkWidget*
gdiff_twopview_new(DiffDir *diffdir, DiffFiles *dfiles, gboolean is_under_dir)
{
	GdiffTwopView *twopview;

	twopview = gtk_type_new(GDIFF_TYPE_TWOPVIEW);

	create_twopane(twopview, diffdir, dfiles);
	create_overview(twopview, dfiles);
	create_vscrollbar(twopview, dfiles);
	create_rmenu(twopview);

	twopview->is_under_dir = is_under_dir;
	
	return GTK_WIDGET(twopview);
}

static void
gdiff_twopview_destroy(GtkObject *object)
{
	GdiffTwopView *twopview;
	GdiffTwoPane *twopane;

	g_return_if_fail(object != NULL);
	g_return_if_fail(GDIFF_IS_TWOPVIEW(object));

	twopview = GDIFF_TWOPVIEW(object);
	twopane = GDIFF_TWOPANE(twopview->twopane);
	gtk_signal_disconnect_by_data(GTK_OBJECT(twopane), twopview);
	
	GTK_OBJECT_CLASS(parent_class)->destroy(object);
}



/** Internal functions **/
static void
gdiff_twopview_set_basepane(GdiffTwopView *twopview, GdiffBasePane *basepane)
{
	g_return_if_fail(twopview != NULL);
	g_return_if_fail(GDIFF_IS_TWOPVIEW(twopview));
	g_return_if_fail(twopview->twopane == NULL);
	
	if (basepane)
		g_return_if_fail(GDIFF_IS_TWOPANE(basepane));

	twopview->twopane = basepane;
}

static void
create_twopane(GdiffTwopView *twopview, DiffDir *diffdir, DiffFiles *dfiles)
{
	GtkWidget *twopane;

	twopane = gdiff_twopane_new(diffdir, dfiles);
	gdiff_twopview_set_basepane(twopview, GDIFF_BASEPANE(twopane));

	gtk_box_pack_start(GTK_BOX(twopview), twopane, TRUE, TRUE, 0);
	gtk_signal_connect(GTK_OBJECT(twopane), "move_diff",
					   GTK_SIGNAL_FUNC(move_diff_cb), twopview);
	gtk_widget_show(twopane);
}

/**
 * create_overview:
 * Create GdiffOverview widget.
 * See "gdiffoverview.[ch]" about GdiffOverview widget.
 **/
static void
create_overview(GdiffTwopView *twopview, DiffFiles *dfiles)
{
	GtkWidget *overview;
	const FileInfo *fi1 = dfiles_get_fileinfo(dfiles, FIRST_FILE, TRUE);
	const FileInfo *fi2 = dfiles_get_fileinfo(dfiles, SECOND_FILE, TRUE);
	GdiffTwoPane *twopane = GDIFF_TWOPANE(twopview->twopane);

	overview = gdiff_overview_new(GTK_TEXT(twopane->text[FIRST_FILE])->vadj,
								  GTK_TEXT(twopane->text[SECOND_FILE])->vadj);
	twopview->overview = GDIFF_OVERVIEW(overview);
	gdiff_overview_set_color1(GDIFF_OVERVIEW(overview),
							  &PANE_PREF(twopane).diff_bg[FIRST_FILE],
							  NULL);
	gdiff_overview_set_color2(GDIFF_OVERVIEW(overview),
							  &PANE_PREF(twopane).diff_bg[SECOND_FILE],
							  NULL);

	/* initialize overview widget */
	if (fi1->nlines != 0 && fi2->nlines != 0) {
		GList *node;/* node of DiffLines list */

		for (node = dfiles->dlines_list; node; node = node->next) {
			const DiffLines *dlines = node->data;
			/* Use local variables for ease to read */
			int begin1 = dlines->between[FIRST_FILE].begin;
			int end1 = dlines->between[FIRST_FILE].end;
			int begin2 = dlines->between[SECOND_FILE].begin;
			int end2 = dlines->between[SECOND_FILE].end;
			
			gdiff_overview_insert_paintrange(GDIFF_OVERVIEW(overview),
											 (gdouble)begin1/fi1->nlines,
											 (gdouble)end1/fi1->nlines,
											 (gdouble)begin2/fi2->nlines,
											 (gdouble)end2/fi2->nlines);
		}
	}
	
	gtk_box_pack_start(GTK_BOX(twopview), overview, FALSE, FALSE, 0);
	gtk_widget_show(overview);
}


/**
 * create_vscrollbar:
 * Create a vertical scrollbar to control both files.
 **/
static void
create_vscrollbar(GdiffTwopView *twopview, DiffFiles *dfiles)
{
#define PAGE_RATIO		10.0	/* XXX hard-coded is good? */
	GtkWidget *vscrollbar;
	GtkAdjustment *vadj;/* adj of vscrollboth */
	int max_nlines;

	max_nlines = dfiles_get_max_nlines(dfiles);
	vadj = GTK_ADJUSTMENT(gtk_adjustment_new(0, 0, max_nlines, 1, max_nlines/PAGE_RATIO, 1));
	vscrollbar = gtk_vscrollbar_new(vadj);
	twopview->vscrollboth = GTK_VSCROLLBAR(vscrollbar);
	gtk_box_pack_start(GTK_BOX(twopview), vscrollbar, FALSE, FALSE, 0);
	gtk_signal_connect(GTK_OBJECT(vadj), "value_changed",
					   GTK_SIGNAL_FUNC(vscrollbar_vchanged), twopview);
	gtk_widget_show(vscrollbar);
}

/**
 * vscrollbar_vchanged:
 * Called when a user changes the value of scrollbar to control both files.
 **/
static void
vscrollbar_vchanged(GtkAdjustment *vadj, gpointer data)
{
	GdiffTwopView *twopview = data;
	GdiffTwoPane *twopane = GDIFF_TWOPANE(twopview->twopane);
	GtkAdjustment *adj1 = GTK_TEXT(twopane->text[FIRST_FILE])->vadj;
	GtkAdjustment *adj2 = GTK_TEXT(twopane->text[SECOND_FILE])->vadj;
	gdouble f1_max = adj1->upper - adj1->lower - adj1->page_size;
	gdouble f2_max = adj2->upper - adj2->lower - adj2->page_size;
	gdouble max_nlines = vadj->upper;
	gdouble value = vadj->value;

	g_assert(max_nlines != 0);
	gtk_adjustment_set_value(adj1, value / max_nlines * f1_max);
	gtk_adjustment_set_value(adj2, value / max_nlines * f2_max);
}


/* right-click menu */
static void
create_rmenu(GdiffTwopView *twopview)
{
	GdiffTwoPane *twopane = GDIFF_TWOPANE(twopview->twopane);
	GtkWidget *rmenu;
	int n;

	for (n = 0; n < NUM_COMPARE_FILES; n++) {
		rmenu = rmenu_create(GTK_WIDGET(twopview), GINT_TO_POINTER(n));
		gnome_popup_menu_attach(rmenu, twopane->text[n], twopview);
	}
}


/**
 * move_diff_cb:
 * Callback for "move_diff" signal of GdiffTwoPane widget.
 * "move_diff" itself is handled by GdiffTwoPane widget.
 * GdiffTwopView widget's responsibility is to take care of vscrollbar
 * and statusbar.
 * Theoretically, it doesn't matter which one I use adj1 or adj2
 * for vscrollbar calculation.
 * I'm using adj1 here.
 * The calculation is reverse of vscrollbar_vchanged()'s one.
 **/
static void
move_diff_cb(GdiffTwoPane *twopane, MoveDiff mv_diff, gpointer data)
{
	GdiffTwopView *twopview = data;
	GtkVScrollbar *vscrollboth = twopview->vscrollboth;
	GtkAdjustment *adj1 = GTK_TEXT(twopane->text[FIRST_FILE])->vadj;
	GtkAdjustment *vadj;/* adj of vscrollboth */
	gdouble max_nlines;
	gdouble f1_max = adj1->upper - adj1->lower - adj1->page_size;

	if (f1_max == 0)
		return;
	
	vadj = gtk_range_get_adjustment(GTK_RANGE(vscrollboth));
	max_nlines = vadj->upper;

	gtk_signal_handler_block_by_func(
		GTK_OBJECT(vadj), GTK_SIGNAL_FUNC(vscrollbar_vchanged), twopview);
	gtk_adjustment_set_value(vadj, adj1->value / f1_max * max_nlines);
	gtk_signal_handler_unblock_by_func(
		GTK_OBJECT(vadj), GTK_SIGNAL_FUNC(vscrollbar_vchanged), twopview);

	update_statusbar(twopview);
}

static void
update_statusbar(GdiffTwopView *twopview)
{
	GdiffBasePane *twopane = GDIFF_BASEPANE(twopview->twopane);
	char *sbar_msg = NULL;/* status-bar message */

	if (twopane->cur_dlines_node) {
		const DiffLines *dlines;
		
		dlines = twopane->cur_dlines_node->data;
		sbar_msg = make_statusbar_msg(
			GDIFF_TWOPVIEW_FILENAME(twopview, FIRST_FILE),
			GDIFF_TWOPVIEW_FILENAME(twopview, SECOND_FILE),
			dlines->between[FIRST_FILE].begin,
			dlines->between[FIRST_FILE].end,
			dlines->between[SECOND_FILE].begin,
			dlines->between[SECOND_FILE].end);
	}
	if (sbar_msg) {
		statusbar_update(twopview->gdwin, sbar_msg);
		g_free(sbar_msg);
	}
}
